(*----------------------------------------------------------------------------*
 *  Direct3D VWPick sample                                                    *
 *  This sample is based on MS DirectX SDK Pick sample demostrating how to    *
 *  "Pick" what triangle is pointed by mouse pointer on rotating object. This *
 *  sample goes further so object is rotated by both WORLD and VIEW matrices  *
 *  (in original sample object was rotating only in VIEW space) AND user can  *
 *  move object in view space by dragging it with right mouse button.         *
 *                                                                            *
 *  Note: This sample requires DirectX 8.1 compatible headers to compile !!!  *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 02-Sep-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit PickUnit;

interface

uses
  Windows, Messages, SysUtils,
  D3DX8, D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil,
  {$IFDEF DXG_COMPAT} DirectXGraphics {$ELSE} Direct3D8 {$ENDIF};

// From 'resource.h'
const
  IDM_TOGGLED3DX                  = 40011;
  IDM_HELP                        = 40012;
  IDM_TOGGLEALLHITS               = 40013;

type
  D3DVERTEX = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    tu, tv: Single;
  end;

const
  D3DFVF_VERTEX = D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1;

type
  TIntersection = record
    dwFace: DWORD;               // mesh face that was intersected
    fBary1, fBary2: Single;      // barycentric coords of intersection
    fDist: Single;               // distance from ray origin to intersection
    tu, tv: Single;              // texture coords of intersection
  end;

// For simplicity's sake, we limit the number of simultaneously intersected 
// triangles to 16
const
  MAX_INTERSECTIONS = 16;

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont: CD3DFont;                        // Font for drawing text
    m_pFontSmall: CD3DFont;                   // Font for drawing text
    m_ArcBall: CD3DArcBall;                   // mouse rotation utility
    m_pObject: CD3DMesh;                      // Object to render
    m_dwNumIntersections: DWORD;              // Number of faces intersected
    m_IntersectionArray: array [0..MAX_INTERSECTIONS-1] of TIntersection; // Intersection info
    m_pVB: IDirect3DVertexBuffer8;            // VB for picked triangles
    m_bShowHelp: BOOL;                        // Whether to show help
    m_bUseD3DX: BOOL;                         // Whether to use D3DXIntersect
    m_bClosestOnly: BOOL;                     // Whether to just get the closest intersection

    // Internal member functions
    function Pick: HResult;
    function IntersectTriangle(const orig, dir: TD3DXVector3;
                               const v0, v1, v2: TD3DXVector3;
                               out t, u, v: Single): BOOL;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
    function ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult; override;

  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;

implementation

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create(0);
  m_strWindowTitle    := 'Pick: D3D Picking Sample';
  m_bUseDepthBuffer   := True;
  m_bShowCursorWhenFullscreen := True;

  m_pFont           := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall      := CD3DFont.Create('Arial', 9, D3DFONT_BOLD);
  m_ArcBall         := CD3DArcBall.Create;
  m_pObject         := CD3DMesh.Create;
  m_dwNumIntersections := 0;
  m_pVB             := nil;
  m_bShowHelp       := False;
  m_bUseD3DX        := True;
  m_bClosestOnly    := False;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Pick()
// Desc: Checks if mouse point hits geometry
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Pick;
type
  TD3DVertexArray = array [0..MaxInt div SizeOf(D3DVERTEX)-1] of D3DVERTEX;
  TD3DXIntersectInfoArray = array [0..MaxInt div SizeOf(TD3DXIntersectInfo)-1] of TD3DXIntersectInfo;
  PD3DXIntersectInfoArray = ^TD3DXIntersectInfoArray;
  TD3DXVector3Array = array [0..0] of TD3DXVector3;
var
  vPickRayDir: TD3DXVector3;
  vPickRayOrig: TD3DXVector3;

  matProj: TD3DXMatrix;
  ptCursor: TPoint;
  v: TD3DXVector3;

  pMesh: ID3DXBaseMesh;
  pVB: IDirect3DVertexBuffer8;
  pIB: IDirect3DIndexBuffer8;

  bHit: BOOL;
  dwFace: DWORD;
  fBary1, fBary2, fDist: Single;

  pBuffer: ID3DXBuffer;
  pIntersectInfoArray: PD3DXIntersectInfoArray;

  iIntersection: DWORD;

  matView, matWorld, m: TD3DXMatrix;

  pVertices: ^TD3DVertexArray;
  pIndices: ^TWordArray;
  dwNumFaces: DWord;

  i: DWord;
  v0, v1, v2: TD3DXVector3;
  pV: ^TD3DVertexArray;

  vThisTri: ^TD3DVertexArray;
  iThisTri: PWordArray;
  pIntersection: ^TIntersection;

  dtu1, dtu2, dtv1, dtv2: Single;
begin
  m_dwNumIntersections:= 0;

  // Get the pick ray from the mouse position
  if GetCapture <> 0 then
  begin
    m_pd3dDevice.GetTransform(D3DTS_PROJECTION, matProj);

    GetCursorPos(ptCursor);
    ScreenToClient(m_hWnd, ptCursor);

    // Compute the vector of the pick ray in screen space
    v.x :=  ( ( ( 2.0 * ptCursor.x ) / m_d3dsdBackBuffer.Width  ) - 1 ) / matProj._11;
    v.y := -( ( ( 2.0 * ptCursor.y ) / m_d3dsdBackBuffer.Height ) - 1 ) / matProj._22;
    v.z :=  1.0;

    // Get the inverse view matrix
    m_pd3dDevice.GetTransform(D3DTS_VIEW, matView);
    m_pd3dDevice.GetTransform(D3DTS_WORLD, matWorld);
    D3DXMatrixMultiply(m, matWorld, matView);
    D3DXMatrixInverse(m, nil, m);

    // Transform the screen space pick ray into 3D space
    vPickRayDir.x  := v.x*m._11 + v.y*m._21 + v.z*m._31;
    vPickRayDir.y  := v.x*m._12 + v.y*m._22 + v.z*m._32;
    vPickRayDir.z  := v.x*m._13 + v.y*m._23 + v.z*m._33;
    vPickRayOrig.x := m._41;
    vPickRayOrig.y := m._42;
    vPickRayOrig.z := m._43;
  end;

  // Get the picked triangle
  if GetCapture <> 0 then
  begin
    pMesh := m_pObject.GetLocalMesh;

    pMesh.GetVertexBuffer(pVB);
    pMesh.GetIndexBuffer(pIB);
    pIB.Lock(0,0, PByte(pIndices), 0);
    pVB.Lock(0,0, PByte(pVertices), 0);

    if m_bUseD3DX then
    begin
      // When calling D3DXIntersect, one can get just the closest intersection and not
      // need to work with a D3DXBUFFER.  Or, to get all intersections between the ray and
      // the mesh, one can use a D3DXBUFFER to receive all intersections.  We show both
      // methods.
      if m_bClosestOnly then
      begin
        // Collect only the closest intersection
        D3DXIntersect(pMesh, vPickRayOrig, vPickRayDir, bHit, @dwFace, @fBary1, @fBary2, @fDist,
            nil, nil);
        if bHit then
        begin
          m_dwNumIntersections := 1;
          m_IntersectionArray[0].dwFace := dwFace;
          m_IntersectionArray[0].fBary1 := fBary1;
          m_IntersectionArray[0].fBary2 := fBary2;
          m_IntersectionArray[0].fDist := fDist;
        end else
        begin
          m_dwNumIntersections := 0;
        end
      end else
      begin
        // Collect all intersections
        pBuffer:= nil;
        Result:= D3DXIntersect(pMesh, vPickRayOrig, vPickRayDir, bHit, nil, nil, nil, nil,
            @pBuffer, @m_dwNumIntersections);
        if FAILED(Result) then Exit;

        if (m_dwNumIntersections > 0) then
        begin
          pIntersectInfoArray := PD3DXIntersectInfoArray(pBuffer.GetBufferPointer);
          if (m_dwNumIntersections > MAX_INTERSECTIONS) then
            m_dwNumIntersections := MAX_INTERSECTIONS;

          for iIntersection:= 0 to m_dwNumIntersections - 1 do
          begin
            m_IntersectionArray[iIntersection].dwFace := pIntersectInfoArray[iIntersection].FaceIndex;
            m_IntersectionArray[iIntersection].fBary1 := pIntersectInfoArray[iIntersection].U;
            m_IntersectionArray[iIntersection].fBary2 := pIntersectInfoArray[iIntersection].V;
            m_IntersectionArray[iIntersection].fDist := pIntersectInfoArray[iIntersection].Dist;
          end;
        end;
      end;
    end else
    begin
      // Not using D3DX
      dwNumFaces := m_pObject.GetLocalMesh.GetNumFaces;
      for i:= 0 to dwNumFaces - 1 do
      begin
        v0 := pVertices[pIndices[3*i+0]].p;
        v1 := pVertices[pIndices[3*i+1]].p;
        v2 := pVertices[pIndices[3*i+2]].p;

        // Check if the pick ray passes through this point
        if IntersectTriangle(vPickRayOrig, vPickRayDir, v0, v1, v2,
                             fDist, fBary1, fBary2) then
        begin
          if (not m_bClosestOnly) or (m_dwNumIntersections = 0) or (fDist < m_IntersectionArray[0].fDist) then
          begin
            if m_bClosestOnly then
              m_dwNumIntersections := 0;
            m_IntersectionArray[m_dwNumIntersections].dwFace := i;
            m_IntersectionArray[m_dwNumIntersections].fBary1 := fBary1;
            m_IntersectionArray[m_dwNumIntersections].fBary2 := fBary2;
            m_IntersectionArray[m_dwNumIntersections].fDist := fDist;
            Inc(m_dwNumIntersections);
            if (m_dwNumIntersections = MAX_INTERSECTIONS) then Break;
          end;
        end;
      end;
    end;

    // Now, for each intersection, add a triangle to m_pVB and compute texture coordinates
    if (m_dwNumIntersections > 0) then
    begin
      m_pVB.Lock(0, 0, PByte(pV), 0);

      for iIntersection:= 0 to m_dwNumIntersections - 1 do
      begin
        pIntersection := @m_IntersectionArray[iIntersection];

        vThisTri := @pV[iIntersection * 3];
        iThisTri := @pIndices[3*pIntersection.dwFace];
        // get vertices hit
        vThisTri[0] := pVertices[iThisTri[0]];
        vThisTri[1] := pVertices[iThisTri[1]];
        vThisTri[2] := pVertices[iThisTri[2]];

        // If all you want is the vertices hit, then you are done.  In this sample, we
        // want to show how to infer texture coordinates as well, using the BaryCentric
        // coordinates supplied by D3DXIntersect
        dtu1 := vThisTri[1].tu - vThisTri[0].tu;
        dtu2 := vThisTri[2].tu - vThisTri[0].tu;
        dtv1 := vThisTri[1].tv - vThisTri[0].tv;
        dtv2 := vThisTri[2].tv - vThisTri[0].tv;
        pIntersection.tu := vThisTri[0].tu + pIntersection.fBary1 * dtu1 + pIntersection.fBary2 * dtu2;
        pIntersection.tv := vThisTri[0].tv + pIntersection.fBary1 * dtv1 + pIntersection.fBary2 * dtv2;
      end;
      m_pVB.Unlock;
    end;


    pVB.Unlock;
    pIB.Unlock;

    pVB:= nil;
    pIB:= nil;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove;
var
  vFromPt:      TD3DXVector3;
  vLookatPt:    TD3DXVector3;
  vUpVec:       TD3DXVector3;

  matView, m: TD3DXMatrix;
begin
  // Rotate the camera about the y-axis
  vFromPt   := D3DXVector3Zero;
  vLookatPt := D3DXVector3Zero;
  vUpVec    := D3DXVector3(0.0, 1.0, 0.0);
  vFromPt.x := -Cos(m_fTime/3.0) * 4.0;
  vFromPt.y := 1.0;
  vFromPt.z :=  Sin(m_fTime/3.0) * 4.0;

  D3DXMatrixLookAtLH(matView, vFromPt, vLookatPt, vUpVec);
  D3DXMatrixMultiply(m, matView, m_ArcBall.GetTranslationMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m);

  D3DXMatrixRotationAxis(m, D3DXVector3(1, 1, 1), m_fTime/1.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render;
var
  ptCursor: TPoint;

  yText: Single;
  iIntersection: DWORD;
  pIntersection: ^TIntersection;
  strBuffer: array [0..89] of Char;
begin
  // Set up the cursor
  GetCursorPos(ptCursor);
  ScreenToClient(m_hWnd, ptCursor);
  m_pd3dDevice.SetCursorPosition(ptCursor.x, ptCursor.y, 0);

  // Check for picked triangles
  Pick;

  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Set render mode to lit, solid triangles
    m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, Integer(TRUE));

    // If a triangle is picked, draw it
    if m_dwNumIntersections <> 0 then
    begin
      // Draw the picked triangle
      m_pd3dDevice.SetVertexShader(D3DFVF_VERTEX);
      m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(D3DVERTEX));
      m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_dwNumIntersections);

      // Set render mode to unlit, wireframe triangles
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
      m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFalse);
    end;

    // Render the mesh
    m_pObject.Render(m_pd3dDevice);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // Output help
    yText := 40.0;
    if m_bShowHelp then
    begin
      m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'F1: Toggle Help');
      yText:= yText + 20.0;
      m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'F2: Change Device');
      yText:= yText + 20.0;
      m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'F3: Toggle D3DX Usage');
      yText:= yText + 20.0;
      m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'F4: Toggle All Hits');
      yText:= yText + 20.0;
    end else
    begin
      m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'Press F1 for Help');
      yText:= yText + 20.0;
    end;

    // Output info
    if m_bUseD3DX
      then m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'Using D3DX')
      else m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'Not Using D3DX');
    yText:= yText + 20.0;

    if m_bClosestOnly
      then m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'Showing First Hit')
      else m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,255,255,255), 'Showing All Hits');
    yText:= yText + 20.0;

    if (m_dwNumIntersections > 0) then
    begin
      for iIntersection:= 0 to m_dwNumIntersections - 1 do
      begin
        pIntersection := @m_IntersectionArray[iIntersection];
        StrFmt(strBuffer, 'Face=%d, tu=%3.02f, tv=%3.02f', [pIntersection.dwFace, pIntersection.tu, pIntersection.tv]);
        m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,0,255,255), strBuffer);
        yText:= yText + 20.0;
      end;
    end else
    begin
      m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,0,255,255), 'Use left mouse button mouse to pick a polygon');
        yText:= yText + 20.0;
      m_pFontSmall.DrawText(2, yText, D3DCOLOR_ARGB(255,0,255,255), 'Use right mouse button mouse to move object');
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects;
var
  dwNumVertices: DWord; 
begin
  // Restore the font
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  if FAILED(m_pObject.Create_(m_pd3dDevice, PChar('Tiger.x'))) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;

  m_pObject.SetFVF(m_pd3dDevice, D3DFVF_VERTEX);

  // Create the vertex buffer
  dwNumVertices := MAX_INTERSECTIONS * 3;
  if FAILED(m_pd3dDevice.CreateVertexBuffer(dwNumVertices*SizeOf(D3DVERTEX),
                                            D3DUSAGE_WRITEONLY, D3DFVF_VERTEX,
                                            D3DPOOL_MANAGED, m_pVB)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
var
  matIdentity: TD3DXMatrix;
  matProj: TD3DXMatrix;
  fAspect: Single;
  mtrl: TD3DMaterial8;
  light: TD3DLight8;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  // Setup the arcball parameters
  m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85);
  m_ArcBall.SetRadius(1.0);

  // Restore device-memory objects for the mesh
  m_pObject.RestoreDeviceObjects(m_pd3dDevice);

  // Set up the textures
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,        iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,        $00444444);

  // Set the world matrix
  D3DXMatrixIdentity(matIdentity);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matIdentity);

  // Set the projection matrix
  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Setup a material
  D3DUtil_InitMaterial(mtrl, 1.0, 1.0, 1.0, 1.0);
  m_pd3dDevice.SetMaterial(mtrl);

  // Set up lighting states
  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, 0.1, -1.0, 0.1);
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;
  m_pObject.InvalidateDeviceObjects;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;
  m_pObject.Destroy_;

  m_pVB:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pFontSmall);
  FreeAndNil(m_ArcBall);
  FreeAndNil(m_pObject);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord;
  Format: TD3DFormat): HResult;
begin
  if dwBehavior and D3DCREATE_PUREDEVICE <> 0 then
  begin
    Result:= E_FAIL; // GetTransform doesn't work on PUREDEVICE
    Exit;
  end;

  // If this is a TnL device, make sure it supports directional lights
  if (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
     (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) then
  begin
    if (pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS = 0) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_ArcBall.HandleMouseMessages(hWnd, uMsg, wParam, lParam);

  // Trap context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result := 0;
    Exit;
  end;
  
  case uMsg of
    WM_LBUTTONDOWN: SetCapture(hWnd); // User pressed left mouse button
    WM_LBUTTONUP:   ReleaseCapture;   // The user released the left mouse button
    WM_COMMAND:
    begin
      if (LOWORD(wParam) = IDM_HELP)          then m_bShowHelp := not m_bShowHelp;
      if (LOWORD(wParam) = IDM_TOGGLED3DX)    then m_bUseD3DX := not m_bUseD3DX;
      if (LOWORD(wParam) = IDM_TOGGLEALLHITS) then m_bClosestOnly := not m_bClosestOnly;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

//-----------------------------------------------------------------------------
// Name: IntersectTriangle()
// Desc: Given a ray origin (orig) and direction (dir), and three vertices of
//       of a triangle, this function returns TRUE and the interpolated texture
//       coordinates if the ray intersects the triangle
//-----------------------------------------------------------------------------
function CMyD3DApplication.IntersectTriangle(const orig, dir: TD3DXVector3;
                             const v0, v1, v2: TD3DXVector3;
                             out t, u, v: Single): BOOL;
var
  edge1: TD3DXVector3;
  edge2: TD3DXVector3;

  pvec: TD3DXVector3;
  det: Single;
  tvec: TD3DXVector3;
  qvec: TD3DXVector3;
  fInvDet: Single;
begin
  // Find vectors for two edges sharing vert0
  D3DXVec3Subtract(edge1, v1, v0);
  D3DXVec3Subtract(edge2, v2, v0);

  // Begin calculating determinant - also used to calculate U parameter
  D3DXVec3Cross(pvec, dir, edge2);

  // If determinant is near zero, ray lies in plane of triangle
  det:= D3DXVec3Dot(edge1, pvec);
  if (det < 0.0001) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Calculate distance from vert0 to ray origin
  D3DXVec3Subtract(tvec, orig, v0);

  // Calculate U parameter and test bounds
  u:= D3DXVec3Dot(tvec, pvec);
  if(u < 0.0) or (u > det) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Prepare to test V parameter
  D3DXVec3Cross(qvec, tvec, edge1);

  // Calculate V parameter and test bounds
  v:= D3DXVec3Dot(dir, qvec);
  if (v < 0.0) or (u + v > det) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Calculate t, scale parameters, ray intersects triangle
  t := D3DXVec3Dot(edge2, qvec);
  fInvDet:= 1.0 / det;
  t:= t * fInvDet;
  u:= u * fInvDet;
  v:= v * fInvDet;

  Result:= True;
end;

end.

